/* Definitions for Intel 386 running FreeBSD with ELF format
   Copyright (C) 1996, 2000, 2002 Free Software Foundation, Inc.
   Contributed by Eric Youngdale.
   Modified for stabs-in-ELF by H.J. Lu.
   Adapted from GNU/Linux version by John Polstra.
   Continued development by David O'Brien <obrien@freebsd.org>

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/* $FreeBSD$ */

#undef  CC1_SPEC
#define CC1_SPEC "%(cc1_cpu) %{profile:-p} \
  %{gline:%{!g:%{!g0:%{!g1:%{!g2: -g1}}}}} \
  "

#undef  ASM_SPEC
#define ASM_SPEC	"%{v*: -v}"

#undef  ASM_FINAL_SPEC
#define ASM_FINAL_SPEC	"%|"

/* Provide a LINK_SPEC appropriate for FreeBSD.  Here we provide support
   for the special GCC options -static and -shared, which allow us to
   link things in one of these three modes by applying the appropriate
   combinations of options at link-time. We like to support here for
   as many of the other GNU linker options as possible. But I don't
   have the time to search for those flags. I am sure how to add
   support for -soname shared_object_name. H.J.

   When the -shared link option is used a final link is not being
   done.  */

#undef	LINK_SPEC
#define LINK_SPEC "\
 %{p:%e`-p' not supported; use `-pg' and gprof(1)} \
    %{Wl,*:%*} \
    %{v:-V} \
    %{assert*} %{R*} %{rpath*} %{defsym*} \
    %{shared:-Bshareable %{h*} %{soname*}} \
    %{!shared: \
      %{!static: \
	%{rdynamic: -export-dynamic} \
	%{!dynamic-linker: -dynamic-linker /usr/libexec/ld-elf.so.1}} \
      %{static:-Bstatic}} \
    %{symbolic:-Bsymbolic}"

/* Provide a STARTFILE_SPEC appropriate for FreeBSD.  Here we add the magical
   crtbegin.o file (see crtstuff.c) which provides part of the support for
   getting C++ file-scope static object constructed before entering `main'.  */

#undef STARTFILE_SPEC
#define STARTFILE_SPEC "\
    %{!shared: \
      %{pg:gcrt1.o%s} \
      %{!pg: \
	%{p:gcrt1.o%s} \
	%{!p:crt1.o%s}}} \
    crti.o%s \
    %{!shared:crtbegin.o%s} \
    %{shared:crtbeginS.o%s}"

/* Provide an ENDFILE_SPEC appropriate for FreeBSD/i386.  Here we tack on our
   own magical crtend.o file (see crtstuff.c) which provides part of the
   support for getting C++ file-scope static object constructed before
   entering `main', followed by the normal "finalizer" file, `crtn.o'.  */

#undef  ENDFILE_SPEC
#define ENDFILE_SPEC "\
    %{!shared:crtend.o%s} \
    %{shared:crtendS.o%s} crtn.o%s"


/************************[  Target stuff  ]***********************************/

/* Define the actual types of some ANSI-mandated types.
   Needs to agree with <machine/ansi.h>.  GCC defaults come from c-decl.c,
   c-common.c, and config/<arch>/<arch>.h.  */

#undef  SIZE_TYPE
#define SIZE_TYPE	(TARGET_64BIT ? "long unsigned int" : "unsigned int")

#undef  PTRDIFF_TYPE
#define PTRDIFF_TYPE	(TARGET_64BIT ? "long int" : "int")

#undef  WCHAR_TYPE_SIZE
#define WCHAR_TYPE_SIZE	(TARGET_64BIT ? 32 : BITS_PER_WORD)

#undef  TARGET_VERSION
#define TARGET_VERSION	fprintf (stderr, " (i386 FreeBSD/ELF)");

#define MASK_PROFILER_EPILOGUE	010000000000

#define TARGET_PROFILER_EPILOGUE	(target_flags & MASK_PROFILER_EPILOGUE)
#define TARGET_ELF			1

#undef	SUBTARGET_SWITCHES
#define SUBTARGET_SWITCHES						\
  { "profiler-epilogue",	 MASK_PROFILER_EPILOGUE, "Function profiler epilogue"}, \
  { "no-profiler-epilogue",	-MASK_PROFILER_EPILOGUE, "No function profiler epilogue"},

/* This goes away when the math emulator is fixed.  */
#undef  TARGET_SUBTARGET_DEFAULT
#define TARGET_SUBTARGET_DEFAULT \
  (MASK_80387 | MASK_IEEE_FP | MASK_FLOAT_RETURNS | MASK_NO_FANCY_MATH_387)

/* Don't default to pcc-struct-return, we want to retain compatibility with
   older gcc versions AND pcc-struct-return is nonreentrant.
   (even though the SVR4 ABI for the i386 says that records and unions are
   returned in memory).  */

#undef  DEFAULT_PCC_STRUCT_RETURN
#define DEFAULT_PCC_STRUCT_RETURN 0

/* Tell final.c that we don't need a label passed to mcount.  */
#define NO_PROFILE_COUNTERS	1

/* Output assembler code to FILE to begin profiling of the current function.
   LABELNO is an optional label.  */

#undef  FUNCTION_PROFILER
#define FUNCTION_PROFILER(FILE, LABELNO)  \
  do {									\
    if (flag_pic)							\
      fprintf ((FILE), "\tcall *.mcount%s@GOT(%%ebx)\n");		\
    else								\
      fprintf ((FILE), "\tcall .mcount\n");				\
  } while (0)

/* Output assembler code to FILE to end profiling of the current function.  */

#undef  FUNCTION_PROFILER_EPILOGUE	/* BDE will need to fix this. */


/************************[  Assembler stuff  ]********************************/

/* Override the default comment-starter of "/" from unix.h.  */
#undef  ASM_COMMENT_START
#define ASM_COMMENT_START "#"

/* Override the default comment-starter of "/APP" from unix.h.  */
#undef  ASM_APP_ON
#define ASM_APP_ON	"#APP\n"
#undef  ASM_APP_OFF
#define ASM_APP_OFF	"#NO_APP\n"

/* XXX:DEO do we still need this override to defaults.h ?? */
/* This is how to output a reference to a user-level label named NAME.  */
#undef  ASM_OUTPUT_LABELREF
#define ASM_OUTPUT_LABELREF(FILE, NAME)					\
  do {									\
    char *_name = (NAME);						\
    /* Hack to avoid writing lots of rtl in				\
       FUNCTION_PROFILER_EPILOGUE ().  */				\
    if (*_name == '.' && strcmp(_name + 1, "mexitcount") == 0)		\
      {									\
	if (flag_pic)							\
	  fprintf ((FILE), "*%s@GOT(%%ebx)", _name);			\
	else								\
	  fprintf ((FILE), "%s", _name);				\
      }									\
    else								\
      fprintf (FILE, "%s", _name);					\
} while (0)

/* This is how to hack on the symbol code of certain relcalcitrant
   symbols to modify their output in output_pic_addr_const ().  */

#undef  ASM_HACK_SYMBOLREF_CODE	/* BDE will need to fix this. */

#undef  ASM_OUTPUT_ALIGN
#define ASM_OUTPUT_ALIGN(FILE, LOG)      				\
  do {					     				\
    if ((LOG)!=0) {							\
      if (in_text_section())						\
	fprintf ((FILE), "\t.p2align %d,0x90\n", (LOG));		\
      else								\
	fprintf ((FILE), "\t.p2align %d\n", (LOG));			\
    }									\
  } while (0)

/* A C statement to output to the stdio stream FILE an assembler
   command to advance the location counter to a multiple of 1<<LOG
   bytes if it is within MAX_SKIP bytes.

   This is used to align code labels according to Intel recommendations.  */

/* XXX configuration of this is broken in the same way as HAVE_GAS_SHF_MERGE,
   but it is easier to fix in an MD way.  */

#ifdef HAVE_GAS_MAX_SKIP_P2ALIGN
#define ASM_OUTPUT_MAX_SKIP_ALIGN(FILE, LOG, MAX_SKIP)			\
  do {									\
    if ((LOG) != 0) {							\
      if ((MAX_SKIP) == 0)						\
	fprintf ((FILE), "\t.p2align %d\n", (LOG));			\
      else								\
	fprintf ((FILE), "\t.p2align %d,,%d\n", (LOG), (MAX_SKIP));	\
    }									\
  } while (0)
#endif


/************************[  Debugger stuff  ]*********************************/

#undef  DBX_REGISTER_NUMBER
#define DBX_REGISTER_NUMBER(n)	(TARGET_64BIT ? dbx64_register_map[n]	\
				: (write_symbols == DWARF2_DEBUG	\
	    			  || write_symbols == DWARF_DEBUG)	\
				  ? svr4_dbx_register_map[(n)]		\
				  : dbx_register_map[(n)])

/* The same functions are used to creating the DWARF2 debug info and C++
   unwind info (except.c).  Regardless of the debug format requested, the
   register numbers used in exception unwinding sections still have to be
   DWARF compatible.  IMO the GCC folks may be abusing the DBX_REGISTER_NUMBER
   macro to mean too much.  */
#define DWARF_FRAME_REGNUM(n)	(TARGET_64BIT ? dbx64_register_map[n]	\
				: svr4_dbx_register_map[(n)])

/* stabs-in-elf has offsets relative to function beginning */
#undef  DBX_OUTPUT_LBRAC
#define DBX_OUTPUT_LBRAC(FILE, NAME)					\
  do {									\
    fprintf (asmfile, "%s %d,0,0,", ASM_STABN_OP, N_LBRAC);		\
    assemble_name (asmfile, buf);					\
        fputc ('-', asmfile);						\
        assemble_name (asmfile,						\
	      	 XSTR (XEXP (DECL_RTL (current_function_decl), 0), 0));	\
    fprintf (asmfile, "\n");						\
  } while (0)

#undef  DBX_OUTPUT_RBRAC
#define DBX_OUTPUT_RBRAC(FILE, NAME)					\
  do {									\
    fprintf (asmfile, "%s %d,0,0,", ASM_STABN_OP, N_RBRAC);		\
    assemble_name (asmfile, buf);					\
        fputc ('-', asmfile);						\
        assemble_name (asmfile,						\
		 XSTR (XEXP (DECL_RTL (current_function_decl), 0), 0));	\
    fprintf (asmfile, "\n");						\
  } while (0)
